/*
Program: MCOUNT.C

Author:  Mike Collins, July 1992


--Description--
Display record count, file size and whether fixed/variable length records.
Show record size (fixed) or largest & smallest record size with count for
each (variable).  All records except last must be delimited by one or two
characters (default = CR+LF) not included in record size.  A file ending
Ctrl_Z (decimal 26) is not included in record size or count.  Tabs can
optionally be expanded to calculate record size.

If specified record delimiter is not found, a message will say so.  If
tab expansion is requested and no tabs were found, a message will say so.


--Usage--
MCOUNT [drive:][path]filename [/Dchar1[,char2]] [/Tn]

filename of ? or /? gets help screen

/D   Record delimiter character(s) in decimal (0-255).
     If using two, separate with comma.  Default = 13,10

/T   Number of spaces to expand tabs (1-99) for calculating record size.
     Default = 1 (no tab expansion)

Above switches are optional and may be upper/lower case.

Example: mcount xyzfile  -is equivalent to-  mcount xyzfile /d13,10 /t1


--Note--
This code is public domain.  Feel free to modify it to suit your needs.

*/

#include <stdlib.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <string.h>

#define  BUF_SIZE 1024             // input file buffer size
#define  CTL_Z    26               // end of file indicator
#define  LF       10               // line feed
#define  CR       13               // carriage return
#define  TAB       9               // tab character
#define  ZERO      0               // null character

void help(int);                    // help function prototype
void errmsg(int, unsigned char *); // err message function prototype
int  getnum(unsigned char**);      // validate number function prototype

main (int argc, char *argv[])
{
   int fh;                       // file handle
   int i;                        // work integer
   int status;                   // # bytes in file read
   int adj = 2;                  // size adjustment for last record
   int tab_adj = 0;              // size adjustment for tab expansion
   long file_size;               // file_size
   long max_size = 0;            // max record size
   long min_size = 2147483647;   // min record size
   long rec_size = 0;            // record size
   long rec_ctr = 0;             // record count
   long max_ctr = 0;             // count of largest variable length records
   long min_ctr = 0;             // count of smallest variable length records
   long nbytes = 0;              // # bytes processed in file
   char rec_form = 'f';          // record form (fixed/variable)
   char eor = 'n';               // end_of_record flag
   char istab = 'n';             // flag to indicate if tabs found
   unsigned char *ptr;           // pointer for command line & file buffer
   unsigned char *FNptr = NULL;  // pointer to filename in command line
   unsigned char tchar1 = CR;    // 1st (or only) record delimiting char
   unsigned char tchar2 = LF;    // 2nd char of record delimiting pair
   unsigned char last_char;      // used for previous char read
   unsigned char filebuf[BUF_SIZE];  // file buffer


/* Check command line (referred to as CL in following comments) for valid
   parameters.  If parameters were seperated by 1 or more spaces, adjust
   the string lengths in the argv[] array to make argv[1] hold all of the
   command line parameters.  Later, re-adjust the length of argv[1] to make
   it hold only the filename.   */

   if((strcmp(argv[1], "?") == 0) || (strcmp(argv[1], "/?") == 0)) help(0);
   if((argc < 2) || (argc > 4)) errmsg(0, " ");     // valid # params?
   if(argc > 2)
   {                                                // if > 1 CL params,
      ptr = argv[1] + strlen(argv[1]);              // combine argv[1]
      *ptr = ' ';                                   // with argv[2]
   }
   if(argc == 4)                                    // if 3 CL params,
   {                                                // append argv[3]
      ptr = argv[1] + strlen(argv[1]);              // to argv[1]
      *ptr = ' ';
   }
   strupr(argv[1]);                         // command line to uppercase
   ptr = strpbrk(argv[1], "/ ");            // look for CL filename delimiter
   if(ptr == NULL) ptr = argv[1] + strlen(argv[1]);
   else FNptr = ptr;                        // save pointer to end of filename

   while (*ptr != ZERO)                     // parse remaining command line
   {
      if(*ptr == ' ') ptr++;
      if(*(ptr++) != '/') errmsg(3, " ");   // check for valid CL delimiter

      if(*ptr == 'D')                       // if D switch found,
      {                                     // get 1st record delimiter char,
         i = getnum(&ptr);                  // verify & put into tchar1
         if((i < 0) || (i > 255)) errmsg(1, " ");
         tchar1 = i;
         if(*ptr == ',')                    // if 2 record delimiter chars,
         {                                  // get 2nd char, verify & put
            i = getnum(&ptr);               // into tchar2
            if((i < 0) || (i > 255)) errmsg(1, " ");
            tchar2 = i;
         }
         else
         {                                  // if only 1 record delimiter
            adj = 1;                        // char, set end-of-record adj
            tchar2 = tchar1;                // factor and move to tchar2
         }
         continue;
      }
      if(*ptr == 'T')                       // if T switch found,
      {                                     // get tab expansion # on CL,
         i = getnum(&ptr);                  // validate and save in tab_adj
         if((i < 1) || (i > 99)) errmsg(2, " ");
         tab_adj = i - 1;
         continue;
      }
      errmsg(3, " ");                       // if we get here, bad delimiter
   }
   // Command line has now been parsed.  Now adjust filename string.
   if(FNptr != NULL) *FNptr = ZERO;         // set argv[1] to filename only
   FNptr = argv[1];                         // FNptr points to filename


   // Open file for read_only and in binary mode.  Error if file not found.
   if((fh = open(FNptr, O_RDONLY | O_BINARY)) == -1) errmsg(4, FNptr);
   file_size = filelength(fh);               // get DOS file size

   while((status = read(fh, filebuf, BUF_SIZE) ) > 0)    // read into buffer
   {
      ptr = filebuf - 1;                     // set buffer pointer for parse
      for (i = 0; i < status; i++)           // parse buffer
      {
         rec_size++;                         // increment record size
         nbytes++;                           // increment # bytes processsed
         if(*(++ptr) == TAB)                 // tab character found?
         {
            rec_size += tab_adj;             // adj rec_size for tab
            istab = 'y';                     // to remember tab found
         }
         if(i > 0) last_char = *(ptr-1);     // to check for delimiting pair
         if(*ptr == tchar2)
         {                                   // if delimiter found,
            if(adj == 1) eor = 'y';          // set end-of-record flag
            else if((rec_size > 1) && (last_char == tchar1)) eor = 'y';
         }
         if((nbytes < file_size) && (eor == 'n')) continue;

         //  End-of-file processing.
         if((nbytes == file_size) && (eor == 'n'))
         {
            // if record delimiter not found, say so and terminate normally
            if(rec_ctr == 0)
            {
               printf("\nDecimal character");
               if(adj == 2) printf(" pair %u %u ",tchar1, tchar2);
               else printf(" %u ",tchar2);
               printf("not found in %s\n", FNptr);
               printf("File size = %lu\n", file_size);
               exit(0);
            }
            // set last record size adjustment for special cases
            if(*ptr != CTL_Z) adj = 0;       // no delimiter or Ctl_Z
            else
            {
               if(rec_size == 1) continue;   // Ctl_Z is only byte
               else adj = 1;                 // adjust for Ctl_Z
            }
         }

         // We get here if
         // 1) Record delimiter encountered or
         // 2) EOF with at least 1 previous record and last record consists
         //    of something more than just Ctrl_Z

         rec_size -= adj;                    // adjust for delimiter size
         rec_ctr++;                          // increment record count
         if((rec_ctr > 1) && (rec_size != max_size)) rec_form = 'v';
         if(rec_size > max_size)             // adj largest record size
         {
            max_size = rec_size;
            max_ctr = 0;
         }
         if(rec_size == max_size) max_ctr++;
         if(rec_size < min_size)             // adj smallest record size
         {
            min_size = rec_size;
            min_ctr = 0;
         }
         if(rec_size == min_size) min_ctr++;
         rec_size = 0;                       // reset record size
         eor = 'n';                          // reset end_of_record flag
      }
      if(nbytes == file_size) break;         // quit if entire file read
      last_char = *(ptr-1);                  // save last char in buffer
   }

   i = close(fh);                            // close input file
   if (i == -1) errmsg(5, FNptr);

   // Were done, now print report and terminate normally
   printf("\n%s record count = %lu\n", FNptr, rec_ctr);
   printf("Record lengths are ");
   if(rec_form == 'v')
   {
      printf("Variable,  File size = %lu\n", file_size);
      printf("Largest record size = %lu,  ", max_size);
      printf("Count = %lu\n", max_ctr);
      printf("Smallest record size = %lu,  ", min_size);
      printf("Count = %lu\n", min_ctr);
   }
   else printf("Fixed at %lu,  File size = %lu\n", max_size, file_size);
   if((tab_adj > 0) && (istab == 'n'))
      printf("No tab characters found in file.\n");
   exit(0);
}

// Print help screen, and exit with errorlevel passed as parameter
void help(int exit_code)
{
   printf("\n\nMCOUNT -- A little file utility by Mike Collins 1992");
   printf("\n\nDisplay record count, file size and whether ");
   printf("fixed/variable length records.\nShow record size (fixed) or ");
   printf("largest & smallest record size with count for\neach ");
   printf("(variable).  All records except last must be delimited ");
   printf("by one or two\nspecified characters (default = CR+LF).  ");
   printf("Tabs can optionally be expanded\nfor calculating record size.");
   printf("\n\nUsage: MCOUNT [drive:][path]filename [/Dchar1[,char2]] [/Tn]");
   printf("\n/D   Record delimiter character(s) in decimal (0-255)");
   printf("\n     If using two, separate with comma.  Default = 13,10");
   printf("\n/T   Number of spaces to expand tabs (1-99) for calculating ");
   printf("record size.\n     Default = 1 (no tab expansion)\n");
   printf("\nExample: mcount xyzfile  -is equivalent to-  ");
   printf("mcount xyzfile /d13,10 /t1\n");
   exit(exit_code);
}

// Print error message, set errorlevel=1 and either exit or get help screen
void errmsg(int err_num, unsigned char *fname)
{
   char *errs[6] =
   {
      "wrong number of parameters",
      "invalid number for record delimiting character",
      "invalid tab expansion number",
      "invalid command line switch",
      "can't find file",
      "on close for file"
   };
   fprintf(stderr,"\n*** ERROR MCOUNT: -- %s %s\n", errs[err_num], fname);
   if(err_num < 4) help(1);                  // get help screen if needed
   exit(1);
}

// Validate numbers on command line via passed pointer.
// Return value is the number or -1 if error.
int getnum(unsigned char **cptr)
{
   unsigned char *cptr1;                     // work pointer
   int j = 1;                                // work integer
   cptr1 = (*cptr) + 1;                      // save passed pointer
   while(j > 0)                              // parse command line string
   {
      (*cptr)++;                             // increment pointer to next char
      if((**cptr >= '0') && (**cptr <= '9')) // is character a valid digit?
      {
         j = 2;                              // show valid digit found
         continue;
      }
      // If a valid digit not found, check for valid command line delimiter
      if((**cptr==ZERO) || (**cptr==' ') || (**cptr=='/') || (**cptr==','))
      {
         if(j == 1) j = -1;                  // error if no valid digits
         break;                              // encountered before delimiter
      }
      j = -1;                                // error if invalid CL delimiter
   }
   if(j != -1) j = atoi(cptr1);              // if number ok, get as integer
   return(j);                                // return number or -1 for error
}
